#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <cuda_runtime.h>
#include <cublas_v2.h>

//#define	ITER	6
#define	ITER	2

#if TYPE_DOUBLE
	#define	TYPE	double
#else
	#define	TYPE	float
#endif

#define USE_CUDA_EVENT_TIMER	0
#define CHECK_ANSWER		1
#include "test-common.h"
#include "test.h"

// macro for the legacy cublas
#define	cublasAlloc(a,b,c)	cudaMalloc( (c),(a)*(b) )
#define	cublasFree(a)		cudaFree( a )

void
warm_up_GPU( cublasHandle_t handle );

int
main(int argc, char *argv[])
{
	TYPE	alpha = (TYPE)3.0;
	TYPE	beta  = (TYPE)0.0;//4.0;
	TYPE	*A, *B, *C, *D;
	TYPE	*dA, *dB, *dC, *dD;
	int	N, M, K;
	int	LDA, LDB, LDC;
	int	i;
	int	blk;
	FILE	*fp;
#if CHECK_ANSWER
	int	verify = 0;
#endif
	int	id = get_gpu_id();


#if __isDOUBLE__
	enum D_FUNCS	func = GEMV_N;
#else
	enum S_FUNCS	func = GEMV_N;
#endif

	char	*func_name = "";

        cudaError_t st;

        if ( cudaSuccess != (st = cudaSetDevice(id)) ) {
           exit(1);
        }

	cublasStatus_t stat;
	cublasHandle_t handle;
	stat = cublasCreate( &handle );

	ASPEN_init( id );

{
	char *argv0 = argv[0];
	if ( strrchr(argv0, '/') != argv0 ) {
		argv0 = strrchr(argv0,'/')+1;
	}
	if ( 0 == strcmp(argv0,TEST_GEMV_N) ) {
		func      = GEMV_N;
		func_name = NAME_GEMV_N;
	}
	if ( 0 == strcmp(argv0,TEST_GEMV_T) ) {
		func      = GEMV_T;
		func_name = NAME_GEMV_T;
	}
	if ( 0 == strcmp(argv0,TEST_SYMV_U) ) {
		func      = SYMV_U;
		func_name = NAME_SYMV_U;
	}
	if ( 0 == strcmp(argv0,TEST_SYMV_L) ) {
		func      = SYMV_L;
		func_name = NAME_SYMV_L;
	}
}

	if ( argc > 1 ) {
		fp=fopen(argv[1], "r");
	} else {
		fp=fopen("IN", "r");
	}
	if ( fp == NULL ) {
		fprintf(stderr,"Cannot open parameter file.\n");
		exit(1);
	}

	fscanf(fp,"%d", &N);

	if ( argc > 2 ) {
		sscanf(argv[2], "%d", &blk);
		printf("% BLK specified=%d\n", blk);
	} else {
		blk = -1; //64;
	}

	{
		verify = 1;
	}

	print_head( func_name );

#if CHECK_ANSWER
	printf("Run in the %s mode\n",
		verify ? "verify" : "non-verify(fast)" );
	fflush(stdout);
#endif
	printf("*****************************************/-->\n");
	fflush(stdout);


	int N_fault=0;

#if 0
	warm_up_GPU( handle );
#endif

int	Itr = 0;
while (1) {
	fscanf(fp,"%d", &N); Itr++;

	if ( N <= 1 ) break;

	LDA = ((N-1)/(256/sizeof(TYPE))+1);
	LDA *= (256/sizeof(TYPE));

	int	M = N;
	int	LDM = ((M-1)/(256/sizeof(TYPE))+1);
	LDM *= (256/sizeof(TYPE));

//	cudaMallocHost((void **)&A, sizeof(TYPE) *(N*LDA));
	cudaMallocHost((void **)&A, sizeof(TYPE) *(M*LDM));
	cudaMallocHost((void **)&B, sizeof(TYPE) *(N*1));
	cudaMallocHost((void **)&C, sizeof(TYPE) *(N*1));
	cudaMallocHost((void **)&D, sizeof(TYPE) *(N*1));

//	cublasAlloc(N*LDA, sizeof(TYPE), (void**)&dA);
	cublasAlloc(M*LDM, sizeof(TYPE), (void**)&dA);
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dB);
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dC);
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dD);

	if(dA==NULL||dB==NULL||dC==NULL||dD==NULL) {
		fprintf(stderr, "# Fail to allocate Dev_array %dx%d, %d*4\n",
			N, LDA, N);
		fprintf(stderr, "# Fail to allocate memory %x %x %x %x\n",
			dA, dB, dC, dD);
		if(dA!=NULL) cublasFree(dA);
		if(dB!=NULL) cublasFree(dB);
		if(dC!=NULL) cublasFree(dC);
		if(dD!=NULL) cublasFree(dD);
		N_fault++;
		goto FFF;
	} else {
		N_fault=0;
	}

	printf("N= %d ", N);
	fflush(stdout);
	fflush(stdout);
	fflush(stdout);


	for(int i=0;i<N;i++) {
		for(int j=0;j<N;j++) {
			A[j+i*LDA]=(TYPE)(0);
		}
	}
	for(i=0;i<N*1;i++) B[i]=(TYPE)1;
	for(i=0;i<N*1;i++) C[i]=(TYPE)0;
	for(i=0;i<N*1;i++) D[i]=(TYPE)0;

#define	BB		(7)

	if (func==GEMV_N) {
		for(int kk=1;kk<=BB;kk++) {
			for(int k=0;k<N;k++) {
				int j=rand()%N; int i=k;
				A[i+j*LDA] += (TYPE)kk*kk*N;
			}
		}
	}
	if (func==GEMV_T) {
		for(int kk=1;kk<=BB;kk++) {
			for(int k=0;k<N;k++) {
				int i=rand()%N; int j=k;
				A[i+j*LDA] += (TYPE)kk*kk*N;
			}
		}
	}
	if (func==SYMV_U||func==SYMV_L) {
		int	*col = (int *)malloc(sizeof(int)*N);
		for(int kk=1;kk<=BB;kk++) {
			for(int k=0;k<N;k++) {
				col[k] = k;
			}
			for(int k=0;k<N;k++) {
				int i1=rand()%N;
				int i2=rand()%N;
				int j=col[i1]; col[i1]=col[i2]; col[i2]=j;
			}
			for(int k=0;k<N;k++) {
				int i=k, j=col[i];
				A[i+j*LDA] += (TYPE)kk*kk*N;
				A[j+i*LDA] += (TYPE)kk*kk*N;
			}
		}
		free(col);
		if (func==SYMV_U) {
		for(int i=0;i<N;i++) {
			for(int j=i+1;j<N;j++) {
				A[j+i*LDA] = (TYPE)0;
			}
		}
		}
		if (func==SYMV_L) {
		for(int i=0;i<N;i++) {
			for(int j=i+1;j<N;j++) {
				A[i+j*LDA] = (TYPE)0;
			}
		}
		}
	}

	for(int k=0;k<N/BB;k++) {
		while ( 1 ) {
		int j=rand()%N;
			if ( B[j] == (TYPE)1 ) {
				B[j] = (TYPE)0;
			break;
			}
		}
	}
	for(int k=0;k<N/BB;k++) {
		while ( 1 ) {
			int j=rand()%N;
			if ( B[j] == (TYPE)1 ) {
				B[j] = (TYPE)(-1);
				break;
			}
		}
	}

#if 0
for(int i=0;i<N;i++){
	for(int j=0;j<N;j++) printf("%g ",A[i+j*LDA]);
	printf("\n");
}
for(int i=0;i<N;i++){
	printf("%g ",B[i]);
}
	printf("\n");
#endif

//	cublasSetMatrix(N, N, sizeof(TYPE), A, LDA, dA, LDA);
if ( argc > 3 ) { ; } else {
	cublasSetMatrix(LDM, M, sizeof(TYPE), A, LDM, dA, LDM);
	cublasSetVector(N, sizeof(TYPE), B, 1, dB, 1);
	cublasSetVector(N, sizeof(TYPE), C, 1, dC, 1);
	cublasSetVector(N, sizeof(TYPE), D, 1, dD, 1);
}

	fflush(stdout);
	cudaThreadSynchronize();

double	tt[1000];
double	t1 = get_wtime();

	for(i=0;i<ITER;i++){

//	do_usleep(5000);

#if USE_CUDA_EVENT_TIMER
	cudaEvent_t	prestart, start, stop;
	float	elapsed_time_ms;

	cudaEventCreate( &prestart );
	cudaEventCreate( &start );
	cudaEventCreate( &stop );
#endif

#if !USE_CUDA_EVENT_TIMER
	cudaThreadSynchronize();
	double	tt1 = get_wtime();
#endif

	int	incb = 1;
	int	incc = 1;
	char	*opt = "U";

#if USE_CUDA_EVENT_TIMER
	cudaEventRecord( prestart, 0 );
	cudaEventSynchronize( prestart );
	cudaEventRecord( start, 0 );
#endif
	switch (func) {
	case GEMV_N:
		ASPEN_GEMV_n (blk, alpha, beta, dA, dB, dC, LDA, N, N);
		break;
	case GEMV_T:
		ASPEN_GEMV_t (blk, alpha, beta, dA, dB, dC, LDA, N, N);
		break;
	case SYMV_U:
		ASPEN_SYMVu (blk, N, dA, LDA, dB, dC, alpha, beta);
		break;
	case SYMV_L:
		ASPEN_SYMVl (blk, N, dA, LDA, dB, dC, alpha, beta);
		break;
	}

#if 1
	cudaError_t cu_err = cudaGetLastError();
	if ( cu_err != cudaSuccess ) {
		printf("CUDA internal ERROR detected\n"); goto HHH;
	}
#endif

#if USE_CUDA_EVENT_TIMER
	cudaEventRecord( stop, 0 );
	cudaEventSynchronize( stop );
	cudaEventElapsedTime( &elapsed_time_ms, start, stop );
	tt[i]=((double)elapsed_time_ms)*1e-3;
#else
	cudaThreadSynchronize();
	double	tt2 = get_wtime();
	tt[i]=tt2-tt1;
#endif

if ( argc > 3 ) goto FIN;

#if CHECK_ANSWER
if(verify){
if(i==0&&N<15000){
	cudaThreadSynchronize();
	do_usleep(100);
	for(int j=0;j<N;j++) { C[j] = (TYPE)0; }
	cublasGetVector(N, sizeof(TYPE), dC, 1, C, 1);
	cudaThreadSynchronize();
	do_usleep(100);
	cublasSetVector(N, sizeof(TYPE), D, 1, dD, 1);
	cudaThreadSynchronize();
	do_usleep(100);
	switch (func) {
        case GEMV_N:
                cublasgemv
                        ( handle, CUBLAS_OP_N, N, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
                break;
        case GEMV_T:
                cublasgemv
                        ( handle, CUBLAS_OP_T, N, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
                break;
        case SYMV_U:
                cublassymv
                        ( handle, CUBLAS_FILL_MODE_UPPER, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
                break;
        case SYMV_L:
                cublassymv
                        ( handle, CUBLAS_FILL_MODE_LOWER, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
                break;
	}
#if 0
	int	jj = cublasGetError();
	if ( jj != 0 ) {
		printf("BLAS ERROR detected\n"); goto HHH;
	}
#endif
	cudaThreadSynchronize();
	do_usleep(100);

	cublasGetVector(N, sizeof(TYPE), dD, 1, D, 1);

	int	kk=N;
	for(int j=0;j<N;j++){
		{
		// check NAN
			int s = isnan(C[j])||isnan(-C[j]);
			if ( s ) { kk=-1; break; }
		}
		if(C[j]!=D[j]){
			kk=j; break;
		}
	}
	if ( kk == -1 ) { goto GGG; }
	if ( kk<N ) printf(    "      ID ERR TRUE / DIFF \n");
	for(int j=kk;j<N;j++){
		if(C[j]!=D[j]){
			printf(" ERROR %d %ld %ld / %g", 
				j, (long)C[j], (long)D[j], C[j]-D[j]);
			printf(" ERROR %d %e %e / %g", 
				j, C[j], D[j], C[j]-D[j]);
			printf("\n");
		}
	}

	cudaThreadSynchronize();
	do_usleep(100);
}
}
#endif

	}
#if 0
	int status = cublasGetError();
	if (status != CUBLAS_STATUS_SUCCESS) {
	        fprintf (stderr, "!!!! kernel execution error.\n");
		goto GGG; //return EXIT_FAILURE;
	}
#endif
EEE:;

double	t2 = get_wtime();
	cudaThreadSynchronize();
double	t3 = get_wtime();

	{
	for(i=0;i<ITER;i++)
	for(int j=i+1;j<ITER;j++)
	if(tt[i]>tt[j]){ double s=tt[i]; tt[i]=tt[j]; tt[j]=s; }
	}

	double	tmax=0.0;
//	for(i=0;i<ITER;i++) tmax=fmax(tt[i],tmax);
	tmax=tt[ITER-1];

	double	ttotal=0.0;
	for(i=0;i<ITER;i++) ttotal+=tt[i];
	ttotal-=tmax;
//double	t = (ttotal)/(ITER-1);
double  t = tt[(ITER-1)/2];

	double	tvar=0.0;
	for(i=0;i<ITER;i++) tvar+=(tt[i]-t)*(tt[i]-t);
	tvar-=(tmax-t)*(tmax-t);
	tvar=sqrt(tvar)/(ITER-1);

	fflush(stdout);
	printf("%g [s] ", t);
	printf("%g GFLOPS ", ((2.0*N)*N/t*1e-9));
	printf("%g ", tvar);
	if ( tvar/t > 2E-2 ) {
		printf("**%g**", tvar/t );
	}
	printf("%d ", LDA);
	fflush(stdout);
	fflush(stdout);
GGG:
	fflush(stdout);

	cublasFree(dA); //printf(":dA");
	cublasFree(dB); //printf(":dB");
	cublasFree(dC); //printf(":dC");
	cublasFree(dD); //printf(":dC");

FFF:
	cudaFreeHost(A); //printf(":A");
	cudaFreeHost(B); //printf(":B");
	cudaFreeHost(C); //printf(":C");
	cudaFreeHost(D); //printf(":D");

	printf("\n");
	fflush(stdout);
	fflush(stdout);

	if ( N_fault > 4 ) break;
	if ( feof(fp) ) break;
}
	cublasDestroy( handle );
	ASPEN_shutdown();
HHH:
        cudaDeviceReset();
	fclose(fp);
FIN:
	;
}

void
warm_up_GPU( cublasHandle_t handle )
{
	TYPE	alpha = (TYPE)3.0;
	TYPE	beta  = (TYPE)0.0;//4.0;
	TYPE	*A, *B, *C, *D;
	TYPE	*dA, *dB, *dC, *dD;

	// Following execution of CUBLAS is for the intention to
	// warm up the GPU temprature with heavy workload !
	int N, LDA;


	N = 9000;
	LDA = ((N-1)/(256/sizeof(TYPE))+1);
	LDA *= (256/sizeof(TYPE));

	cudaMallocHost((void **)&A, sizeof(TYPE) *(N*LDA));
#if 0
	cudaMallocHost((void **)&B, sizeof(TYPE) *(N*1));
	cudaMallocHost((void **)&C, sizeof(TYPE) *(N*1));
	cudaMallocHost((void **)&D, sizeof(TYPE) *(N*1));
#endif

	cublasAlloc(N*LDA, sizeof(TYPE), (void**)&dA);
#if 0
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dB);
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dC);
	cublasAlloc(N*1, sizeof(TYPE), (void**)&dD);
#endif

#if 0
	for(int i=0;i<N;i++) {
		for(int j=0;j<N;j++) {
			A[j+i*LDA]=(TYPE)(0);
		}
	}
	for(int i=0;i<N*1;i++) B[i]=(TYPE)1;
	for(int i=0;i<N*1;i++) C[i]=(TYPE)0;
	for(int i=0;i<N*1;i++) D[i]=(TYPE)0;

	for(int kk=1;kk<=BB;kk++) {
		for(int k=0;k<N;k++) {
			int i=rand()%N; int j=k;
			A[i+j*LDA] += (TYPE)kk*kk*N;
		}
	}

	for(int k=0;k<N/BB;k++) {
		while ( 1 ) {
		int j=rand()%N;
			if ( B[j] == (TYPE)1 ) {
				B[j] = (TYPE)0;
			break;
			}
		}
	}
	for(int k=0;k<N/BB;k++) {
		while ( 1 ) {
			int j=rand()%N;
			if ( B[j] == (TYPE)1 ) {
				B[j] = (TYPE)(-1);
				break;
			}
		}
	}
#endif
	cublasSetMatrix(N, N, sizeof(TYPE), A, LDA, dA, LDA);
#if 0
	cublasSetVector(N, sizeof(TYPE), B, 1, dB, 1);
	cublasSetVector(N, sizeof(TYPE), C, 1, dC, 1);
	cublasSetVector(N, sizeof(TYPE), D, 1, dD, 1);
#endif

#if 0
	for(int i=0;i<3;i++){
	  cudaThreadSynchronize();
          cublasgemv
	  ( handle, CUBLAS_OP_T, N, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
	  cudaThreadSynchronize();
	}
#endif

	cublasFree(dA); //printf(":dA");
#if 0
	cublasFree(dB); //printf(":dB");
	cublasFree(dC); //printf(":dC");
	cublasFree(dD); //printf(":dC");
#endif

	cudaFreeHost(A); //printf(":A");
#if 0
	cudaFreeHost(B); //printf(":B");
	cudaFreeHost(C); //printf(":C");
	cudaFreeHost(D); //printf(":D");
#endif
	  cudaThreadSynchronize();
}

