#!/bin/bash

_min_ ( ) {
	x=$1; y=$2
	if [ $x -lt $y ]; then
		z=$x
	else
		z=$y
	fi
	echo $z
}

_max_ ( ) {
	x=$1; y=$2
	if [ $x -gt $y ]; then
		z=$x
	else
		z=$y
	fi
	echo $z
}

_ceil_ ( ) {
	x=$1; y=$2
	M=`expr $x "-" 1`
	M=`expr $M "/" $y`
	M=`expr $M "+" 1`
	z=`expr $M "*" $y`
	echo $z
}

_floor_ ( ) {
	x=$1; y=$2
	M=`expr $x "/" $y`
	z=`expr $M "*" $y`
	echo $z
}


# ============================================================

#	BX  Num of the Threads / ThreadBlock
#	CX  Num of the Registers / Thread
#	MX  [B] Minimum spend of Shared Memory / ThreadBlock
#	CG  Compute Capability Generation (100, 200, 300, 500)
#
# ============================================================

	BX=$1
	CX=$2
	MX=$3
	CG=$4

# ============================================================

	SX=`expr 48 "*" 1024`
	WARP=32

	if [ $CG -eq 130 ]; then
	# for Tesla
		MAX_REGS=16384
		MAX_BLK=8
		MAX_WARPS=32
		SHMEM_UNIT_SIZE=512
		REG_UNIT_SIZE=512
		GRANURARITY=2
		SX_BLOCK=1
	fi
	if [ $CG -eq 200 ]; then
	# for Fermi
		MAX_REGS=32768
		MAX_BLK=8
		MAX_WARPS=48
		SHMEM_UNIT_SIZE=128
		REG_UNIT_SIZE=64
		for RS in 21 22 29 30 37 38 45 46
		do
			if [ $CX -eq $RS ]; then
				REG_UNIT_SIZE=128
			fi
		done
		GRANURARITY=2
		SX_BLOCK=1
	fi
	if [ $CG -eq 300 ]; then
	# for Kepler
		MAX_REGS=65536
		MAX_BLK=16
		MAX_WARPS=64
		SHMEM_UNIT_SIZE=256
		REG_UNIT_SIZE=256
		GRANURARITY=4
		SX_BLOCK=1
	fi
	if [ $CG -eq 350 ]; then
	# for Kepler2
		MAX_REGS=65536
		MAX_BLK=16
		MAX_WARPS=64
		SHMEM_UNIT_SIZE=256
		REG_UNIT_SIZE=256
		GRANURARITY=4
		SX_BLOCK=1
	fi
	if [ $CG -eq 500 ]; then
	# for Maxwell
		MAX_REGS=65536
		MAX_BLK=32
		MAX_WARPS=64
		SHMEM_UNIT_SIZE=256
		REG_UNIT_SIZE=256
		GRANURARITY=4
#		SX_BLOCK=4
		SX_BLOCK=1
	fi
	if [ $CG -eq 520 ]; then
	# for Maxwell2
		MAX_REGS=65536
		MAX_BLK=32
		MAX_WARPS=64
		SHMEM_UNIT_SIZE=256
		REG_UNIT_SIZE=256
		GRANURARITY=4
#		SX_BLOCK=4
		SX_BLOCK=1
	fi

	if [ $SX -lt $MX ]; then
		echo '0(0)'
		exit 0
	fi

# ============================================================

	X0=`expr $SX "/" $SHMEM_UNIT_SIZE`

	M=`expr $MX "-" 1`
	M=`expr $M "/" $SHMEM_UNIT_SIZE`
	X1=`expr $M "+" 1`

# ============================================================

	Y0=`expr $CX "*" $WARP`

	Y1=`_ceil_ $Y0 $REG_UNIT_SIZE`

	M=`expr $BX "-" 1`
	M=`expr $M "/" $WARP`
	Y3=`expr $M "+" 1`

	if [ $CG -lt 200 ]; then

		Y2=`_ceil_ $Y3 2`
		M=`expr $Y2 "*" $Y0`
		Y4=`_ceil_ $M $REG_UNIT_SIZE`
		Y5=$MAX_REGS

	else

		M=`expr $MAX_REGS "/" $SX_BLOCK`
		M=`expr $M "/" $Y1`
		Y2=`expr $M "*" $SX_BLOCK`
		Y4=$Y3
		Y5=`_floor_ $Y2 $GRANURARITY`

	fi

# ============================================================

	M1=`expr $Y5 "/" $Y4`
	M2=`expr $X0 "/" $X1`

	M=`expr $MAX_WARPS "/" $Y3`
	M3=`_min_ $M $MAX_BLK`

	M=`_min_ $M1 $M2`
	MM=`_min_ $M $M3`

# ============================================================

	M=`expr $MM "*" $Y3`
	M=`expr $M  "*" 100`

	M1=$MAX_WARPS
	OO=`expr $M "/" $M1`

	M_PATTERN=$MM"("$OO")"

	MM=`expr $MM "-" 1`
	while [ $MM -gt 0 ]
	do
		M6=`expr $X0 "/" $MM`
		M=`expr $MM "+" 1`
		M=`expr $M "*" $M6`
		if [ $M -gt $X0 ]; then
			M=`expr $MM "*" $Y3`
			M=`expr $M  "*" 100`

			M1=$MAX_WARPS
			OO=`expr $M "/" $M1`

			M_PATTERN=$MM"("$OO") "$M_PATTERN
		fi
		MM=`expr $MM "-" 1`
	done

# ============================================================

	echo $M_PATTERN

# ============================================================

exit 0

